package str.classtree.atermify;

import java.util.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.Repository;
import org.apache.bcel.Constants;
import org.apache.bcel.util.InstructionFinder;
import str.bcel.EmptyAbstractionVisitor;
import aterm.ATerm;
import aterm.ATermList;
import static str.classtree.atermify.ATermifyUtil.*;
import str.aterm.ATermifier;
import static str.aterm.ATermUtil.*;

public class InstructionATermifier implements ATermifier<InstructionHandle>
{
  private ConstantPoolGen _cpg;

  private ATermifier<Type> _ftype;
  private ATermifier<Constant> _fconst;
  
  public InstructionATermifier(ConstantPoolGen cpg)
  {
    super();
    _cpg = cpg;
    _ftype = TypeATermifier.SHARED;
    _fconst = new GenConstantATermifier(cpg);
  }
  
  public ATerm convert(InstructionHandle handle)
  {
    Visitor visitor = new Visitor(handle);

    try
    {
      handle.getInstruction().accept(visitor);
    }
    catch(RuntimeException exc)
    {
      System.err.println("class2aterm: error during conversion of " + handle.toString());
      throw exc;
    }

    ATerm result = visitor.getResult();
    if(result == null)
    {
      throw new IllegalArgumentException("Cannot convert instruction : " + handle.getInstruction().toString());
    }
      
    return visitor.getResult();
  }

  private class Visitor extends EmptyAbstractionVisitor
  {
    private ATerm _result;
    InstructionHandle _handle;

    public Visitor(InstructionHandle handle)
    {
      super();
      _handle = handle;
    }

    public ATerm getResult()
    {
      return _result;
    }

    /**
     * Utils
     */

    private int getInstructionCountTo(InstructionHandle target)
    {
      int count = 0;
      
      InstructionHandle current = _handle;
      while(current != target && current != null)
      {
	count--;
	current = current.getPrev();
      }

      if(current == null)
      {
	count = 0;

	current = _handle;
	while(current != target && current != null)
	{
	  count++;
	  current = current.getNext();
	}

	if(current == null)
	{
	  throw new IllegalArgumentException("Branch target not found in instruction list." + count);
	}
      }

      return count;
    }

    private ATerm buildFieldRef(FieldInstruction instr)
    {
      return
	appl("FieldRef",
	  appl("Class", astring(instr.getClassName(_cpg))),
	  appl("Name", astring(instr.getFieldName(_cpg))),
	  buildFieldDescriptor(instr.getFieldType(_cpg))
	);
    }

    private ATerm buildMethodRef(InvokeInstruction instr)
    {
      return
	appl("MethodRef",
	  appl("Class", astring(instr.getClassName(_cpg))),
	  appl("Name", astring(instr.getMethodName(_cpg))),
	  buildMethodDescriptor(instr.getReturnType(_cpg), instr.getArgumentTypes(_cpg))
	);
    }

    private ATerm buildInterfaceMethodRef(InvokeInstruction instr)
    {
      return
	appl("InterfaceMethodRef",
	  appl("Interface", astring(instr.getClassName(_cpg))),
	  appl("Name", astring(instr.getMethodName(_cpg))),
	  buildMethodDescriptor(instr.getReturnType(_cpg), instr.getArgumentTypes(_cpg))
	);
    }

    /**
     * Pop Instructions
     */
    public void visitPOP(POP obj)  
    {       
      _result = appl("POP");
    }

    public void visitPOP2(POP2 obj)  
    {       
      _result = appl("POP2");
    }

    /**
     * Constant Push Instructions
     */
    public void visitBIPUSH(BIPUSH instr)  
    {       
      _result = appl("BIPUSH", appl("Int", astring(instr.getValue().toString())));
    }

    public void visitSIPUSH(SIPUSH instr)  
    {       
      _result = appl("SIPUSH", appl("Int", astring(instr.getValue().toString())));
    }

    public void visitICONST(ICONST obj)  
    {       
      _result = _factory.make(obj.getName().toUpperCase());
    }

    public void visitLCONST(LCONST obj)  
    {       
      _result = _factory.make(obj.getName().toUpperCase());
    }

    public void visitDCONST(DCONST obj)  
    {       
      _result = _factory.make(obj.getName().toUpperCase());
    }

    public void visitFCONST(FCONST obj)  
    {       
      _result = _factory.make(obj.getName().toUpperCase());
    }

    public void visitACONST_NULL(ACONST_NULL obj) 
    {       
      _result = _factory.make("ACONST_NULL");
    }

    public void visitLDC(LDC instr)
    {
      Constant c = _cpg.getConstantPool().getConstant(instr.getIndex());
      ATerm value = _fconst.convert(c);

      if(instr instanceof LDC_W)
      {
        _result = appl("LDC_W", value);
      }
      else
      {
	_result = appl("LDC", value);
      }
    }

    public void visitLDC2_W(LDC2_W instr)
    {
      Constant c = _cpg.getConstantPool().getConstant(instr.getIndex());
      _result = appl("LDC2_W", _fconst.convert(c));
    }

    public void visitNEW(NEW instr)
    {
      _result = appl("NEW", _ftype.convert(instr.getLoadClassType(_cpg)));
    }

    /**
     * DUP Instructions
     */

    public void visitDUP(DUP obj)
    {
      _result = appl("DUP");
    }

    public void visitDUP_X1(DUP_X1 instr)
    {
      _result = appl("DUP_X1");
    }

    public void visitDUP_X2(DUP_X2 instr)
    {
      _result = appl("DUP_X2");
    }

    public void visitDUP2(DUP2 instr)
    {
      _result = appl("DUP2");
    }

    public void visitDUP2_X1(DUP2_X1 instr)
    {
      _result = appl("DUP2_X1");
    }

    public void visitDUP2_X2(DUP2_X2 instr)
    {
      _result = appl("DUP2_X2");
    }

    /**
     * Misc stack instructions
     */

    public void visitSWAP(SWAP instr)
    {
      _result = appl("SWAP");
    }

    /**
     * Invoke Instructions
     */

    public void visitINVOKESPECIAL(INVOKESPECIAL instr)
    {
      _result = appl("INVOKESPECIAL", buildMethodRef(instr));
    }

    public void visitINVOKEVIRTUAL(INVOKEVIRTUAL instr)
    {
      _result = appl("INVOKEVIRTUAL", buildMethodRef(instr));
    }

    public void visitINVOKEINTERFACE(INVOKEINTERFACE instr)
    {
      _result = appl("INVOKEINTERFACE", buildInterfaceMethodRef(instr));
    }

    public void visitINVOKESTATIC(INVOKESTATIC instr)
    {
      _result = appl("INVOKESTATIC", buildMethodRef(instr));
    }

    /**
     * Static fields
     */

    public void visitGETSTATIC(GETSTATIC instr)
    {
      _result = appl("GETSTATIC", buildFieldRef(instr));
    }

    public void visitPUTSTATIC(PUTSTATIC instr)
    {
      _result = appl("PUTSTATIC", buildFieldRef(instr));
    }

    /**
     * Instance Fields
     */
    public void visitGETFIELD(GETFIELD instr)
    {
      _result = appl("GETFIELD", buildFieldRef(instr));
    }

    public void visitPUTFIELD(PUTFIELD instr)
    {
      _result = appl("PUTFIELD", buildFieldRef(instr));
    }

    /**
     * Load local variables
     */

    public void visitILOAD(ILOAD instr)  
    {       
      _result = appl("ILOAD", aint(instr.getIndex()));
    }

    public void visitLLOAD(LLOAD instr)  
    {   
      _result = appl("LLOAD", aint(instr.getIndex()));
    }
  
    public void visitFLOAD(FLOAD instr)  
    {       
      _result = appl("FLOAD", aint(instr.getIndex()));
    }
  
    public void visitDLOAD(DLOAD instr)  
    {       
      _result = appl("DLOAD", aint(instr.getIndex()));
    }

    public void visitALOAD(ALOAD instr)  
    {
      _result = appl("ALOAD", aint(instr.getIndex()));
    }

    /**
     * Store local variables
     */
    public void visitIINC(IINC instr)  
    {       
      _result = appl("IINC", aint(instr.getIndex()), appl("Int", astring(String.valueOf(instr.getIncrement()))));
    }

    public void visitISTORE(ISTORE instr)  
    {       
      _result = appl("ISTORE", aint(instr.getIndex()));
    }

    public void visitLSTORE(LSTORE instr)  
    {   
      _result = appl("LSTORE", aint(instr.getIndex()));
    }
  
    public void visitFSTORE(FSTORE instr)  
    {       
      _result = appl("FSTORE", aint(instr.getIndex()));
    }
  
    public void visitDSTORE(DSTORE instr)  
    {       
      _result = appl("DSTORE", aint(instr.getIndex()));
    }

    public void visitASTORE(ASTORE instr)  
    {
      _result = appl("ASTORE", aint(instr.getIndex()));
    }

    /**
     * Arithmetic Instructions
     */

    public void visitDADD(DADD instr) { _result = appl("DADD"); }
    public void visitDDIV(DDIV instr) { _result = appl("DDIV"); }
    public void visitDMUL(DMUL instr) { _result = appl("DMUL"); }
    public void visitDNEG(DNEG instr) { _result = appl("DNEG"); }
    public void visitDREM(DREM instr) { _result = appl("DREM"); }
    public void visitDSUB(DSUB instr) { _result = appl("DSUB"); }

    public void visitFADD(FADD instr) { _result = appl("FADD"); }
    public void visitFDIV(FDIV instr) { _result = appl("FDIV"); }
    public void visitFMUL(FMUL instr) { _result = appl("FMUL"); }
    public void visitFNEG(FNEG instr) { _result = appl("FNEG"); }
    public void visitFREM(FREM instr) { _result = appl("FREM"); }
    public void visitFSUB(FSUB instr) { _result = appl("FSUB"); }

    public void visitIADD( IADD  instr) { _result = appl("IADD");  }
    public void visitIAND( IAND  instr) { _result = appl("IAND");  }
    public void visitIDIV( IDIV  instr) { _result = appl("IDIV");  }
    public void visitIMUL( IMUL  instr) { _result = appl("IMUL");  }
    public void visitINEG( INEG  instr) { _result = appl("INEG");  }
    public void visitIOR(  IOR   instr) { _result = appl("IOR");   }
    public void visitIREM( IREM  instr) { _result = appl("IREM");  }
    public void visitISHL( ISHL  instr) { _result = appl("ISHL");  }
    public void visitISHR( ISHR  instr) { _result = appl("ISHR");  }
    public void visitISUB( ISUB  instr) { _result = appl("ISUB");  }
    public void visitIUSHR(IUSHR instr) { _result = appl("IUSHR"); }
    public void visitIXOR( IXOR  istr)  { _result = appl("IXOR");  }

    public void visitLADD( LADD  instr) { _result = appl("LADD");  }
    public void visitLAND( LAND  instr) { _result = appl("LAND");  }
    public void visitLDIV( LDIV  instr) { _result = appl("LDIV");  }
    public void visitLMUL( LMUL  instr) { _result = appl("LMUL");  }
    public void visitLNEG( LNEG  instr) { _result = appl("LNEG");  }
    public void visitLOR(  LOR   instr) { _result = appl("LOR");   }
    public void visitLREM( LREM  instr) { _result = appl("LREM");  }
    public void visitLSHL( LSHL  instr) { _result = appl("LSHL");  }
    public void visitLSHR( LSHR  instr) { _result = appl("LSHR");  }
    public void visitLSUB( LSUB  instr) { _result = appl("LSUB");  }
    public void visitLUSHR(LUSHR instr) { _result = appl("LUSHR"); }
    public void visitLXOR( LXOR  istr)  { _result = appl("LXOR");  }

    /**
     * Comparison
     */
    public void visitDCMPG(DCMPG obj) { _result = appl("DCMPG");  }
    public void visitDCMPL(DCMPL obj) { _result = appl("DCMPL");  }

    public void visitFCMPG(FCMPG obj) { _result = appl("FCMPG");  }
    public void visitFCMPL(FCMPL obj) { _result = appl("FCMPL");  }

    public void visitLCMP(LCMP obj)   { _result = appl("LCMP");  }

    /**
     * If
     */
    public void visitIF_ACMPEQ(IF_ACMPEQ instr)
    {
      _result = appl("IF_ACMPEQ", buildIf(instr));
    }

    public void visitIF_ACMPNE(IF_ACMPNE instr)
    {
      _result = appl("IF_ACMPNE", buildIf(instr));
    }

    public void visitIF_ICMPEQ(IF_ICMPEQ instr)
    {
      _result = appl("IF_ICMPEQ", buildIf(instr));
    }

    public void visitIF_ICMPGE(IF_ICMPGE instr)
    {
      _result = appl("IF_ICMPGE", buildIf(instr));
    }

    public void visitIF_ICMPGT(IF_ICMPGT instr)
    {
      _result = appl("IF_ICMPGT", buildIf(instr));
    }

    public void visitIF_ICMPLE(IF_ICMPLE instr)
    {
      _result = appl("IF_ICMPLE", buildIf(instr));
    }

    public void visitIF_ICMPLT(IF_ICMPLT instr)
    {
      _result = appl("IF_ICMPLT", buildIf(instr));
    }

    public void visitIF_ICMPNE(IF_ICMPNE instr)
    {
      _result = appl("IF_ICMPNE", buildIf(instr));
    }

    public void visitIFEQ(IFEQ instr)
    {
      _result = appl("IFEQ", buildIf(instr));
    }

    public void visitIFGE(IFGE instr)
    {
      _result = appl("IFGE", buildIf(instr));
    }

    public void visitIFGT(IFGT instr)
    {
      _result = appl("IFEQ", buildIf(instr));
    }

    public void visitIFLE(IFLE instr)
    {
      _result = appl("IFLE", buildIf(instr));
    }

    public void visitIFLT(IFLT instr)
    {
      _result = appl("IFLT", buildIf(instr));
    }

    public void visitIFNE(IFNE instr)
    {
      _result = appl("IFNE", buildIf(instr));
    }

    public void visitIFNONNULL(IFNONNULL instr)
    {
      _result = appl("IFNONNULL", buildIf(instr));
    }

    public void visitIFNULL(IFNULL instr)
    {
      _result = appl("IFNULL", buildIf(instr));
    }

    private ATerm buildIf(IfInstruction instr)
    {
      return appl("Count", aint(getInstructionCountTo(instr.getTarget())));
    }
  
    /**
     * Conversion Instructions
     */

    public void visitD2F(D2F instr) { _result = appl("D2F"); }
    public void visitD2I(D2I instr) { _result = appl("D2I"); }
    public void visitD2L(D2L instr) { _result = appl("D2L"); }
    public void visitF2D(F2D instr) { _result = appl("F2D"); }
    public void visitF2I(F2I instr) { _result = appl("F2I"); }
    public void visitF2L(F2L instr) { _result = appl("F2L"); }
    public void visitI2B(I2B instr) { _result = appl("I2B"); }
    public void visitI2C(I2C instr) { _result = appl("I2C"); }
    public void visitI2D(I2D instr) { _result = appl("I2D"); }
    public void visitI2F(I2F instr) { _result = appl("I2F"); }
    public void visitI2L(I2L instr) { _result = appl("I2L"); }
    public void visitI2S(I2S instr) { _result = appl("I2S"); }
    public void visitL2D(L2D instr) { _result = appl("L2D"); }
    public void visitL2F(L2F instr) { _result = appl("L2F"); }
    public void visitL2I(L2I instr) { _result = appl("L2I"); }


    public void visitCHECKCAST(CHECKCAST instr)
    {
      _result = appl("CHECKCAST", _ftype.convert(instr.getType(_cpg)));
    }

    public void visitINSTANCEOF(INSTANCEOF instr)
    {
      _result = appl("INSTANCEOF", _ftype.convert(instr.getType(_cpg)));
    }

    /**
     * Return instructions
     */

    public void visitARETURN(ARETURN instr) { _result = appl("ARETURN"); }
    public void visitDRETURN(DRETURN instr) { _result = appl("DRETURN"); }
    public void visitFRETURN(FRETURN instr) { _result = appl("FRETURN"); }
    public void visitIRETURN(IRETURN instr) { _result = appl("IRETURN"); }
    public void visitLRETURN(LRETURN instr) { _result = appl("LRETURN"); }
    public void visitRETURN(RETURN instr)   { _result = appl("RETURN");  }

    /**
     * Array Creation
     */

    public void visitNEWARRAY(NEWARRAY instr)  
    {       
      _result = appl("NEWARRAY", _ftype.convert(((ArrayType)instr.getType()).getElementType()));
    }

    public void visitANEWARRAY(ANEWARRAY instr)  
    {
      _result = appl("ANEWARRAY", _ftype.convert(instr.getType(_cpg)));
    }

    public void visitMULTIANEWARRAY(MULTIANEWARRAY instr)
    {       
      _result = appl("MULTIANEWARRAY", _ftype.convert(instr.getType(_cpg)), appl("Int", astring(String.valueOf(instr.getDimensions()))));
    }

    /**
     * Array Length
     */

    public void visitARRAYLENGTH(ARRAYLENGTH instr)  
    {       
      _result = appl("ARRAYLENGTH");
    }

    /**
     * Array Store Instructions
     */
    public void visitIASTORE(IASTORE instr)  
    {       
      _result = appl("IASTORE");
    }

    public void visitLASTORE(LASTORE instr)  
    {   
      _result = appl("LASTORE");
    }

    public void visitBASTORE(BASTORE instr)  
    {       
      _result = appl("BASTORE");
    }

    public void visitCASTORE(CASTORE instr)  
    {       
      _result = appl("CASTORE");
    }

    public void visitSASTORE(SASTORE instr)  
    {       
      _result = appl("SASTORE");
    }
  
    public void visitFASTORE(FASTORE instr)  
    {       
      _result = appl("FASTORE");
    }
  
    public void visitDASTORE(DASTORE instr)  
    {       
      _result = appl("DASTORE");
    }

    public void visitAASTORE(AASTORE instr)  
    {
      _result = appl("AASTORE");
    }

    /**
     * Array Load Instructions
     */

    public void visitIALOAD(IALOAD instr)  
    {       
      _result = appl("IALOAD");
    }

    public void visitLALOAD(LALOAD instr)  
    {   
      _result = appl("LALOAD");
    }

    public void visitBALOAD(BALOAD instr)  
    {       
      _result = appl("BALOAD");
    }

    public void visitCALOAD(CALOAD instr)  
    {       
      _result = appl("CALOAD");
    }

    public void visitSALOAD(SALOAD instr)  
    {       
      _result = appl("SALOAD");
    }
  
    public void visitFALOAD(FALOAD instr)  
    {       
      _result = appl("FALOAD");
    }
  
    public void visitDALOAD(DALOAD instr)  
    {       
      _result = appl("DALOAD");
    }

    public void visitAALOAD(AALOAD instr)  
    {
      _result = appl("AALOAD");
    }

    /**
     * Exception Instructions
     */

    public void visitATHROW(ATHROW obj)  
    {       
      _result = appl("ATHROW");
    }

    public void visitGOTO(GOTO instr)  
    {       
      _result = appl("GOTO", appl("Count", aint(getInstructionCountTo(instr.getTarget()))));
    }

    public void visitGOTO_W(GOTO_W instr)  
    {       
      _result = appl("GOTO_W", appl("Count", aint(getInstructionCountTo(instr.getTarget()))));
    }

    public void visitJSR(JSR instr)  
    {       
      _result = appl("JSR", appl("Count", aint(getInstructionCountTo(instr.getTarget()))));
    }

    public void visitJSR_W(JSR_W instr)  
    {       
      _result = appl("JSR_W", appl("Count", aint(getInstructionCountTo(instr.getTarget()))));
    }

    public void visitRET(RET instr)
    {
      _result = appl("RET", aint(instr.getIndex()));
    }

    /**
     * Concurrency Instructions
     */

    public void visitMONITORENTER(MONITORENTER obj)  
    {       
      _result = appl("MONITORENTER");
    }
  
    public void visitMONITOREXIT(MONITOREXIT obj)  
    {       
      _result = appl("MONITOREXIT");
    }

    /**
     * Switch
     */

    public void visitLOOKUPSWITCH(LOOKUPSWITCH instr)
    {
      int[] matches = instr.getMatchs();
      InstructionHandle[] targets = instr.getTargets();
  
      ATermList npairs = _factory.makeList();
      for(int i = targets.length - 1; i >= 0; i--)
      {
	npairs = npairs.insert(appl("NPAIR", aint(matches[i]), appl("Count", aint(getInstructionCountTo(targets[i])))));
      }

      _result = appl("LOOKUPSWITCH",
	appl("Count", aint(getInstructionCountTo(instr.getTarget()))),
	npairs);
    }

    public void visitTABLESWITCH(TABLESWITCH instr)
    {
      int[] matches = instr.getMatchs();
      InstructionHandle[] targets = instr.getTargets();
      
      ATermList jumps = _factory.makeList();
      for(int i = targets.length - 1; i >= 0; i--)
      {
	jumps = jumps.insert(appl("Jump", appl("Count", aint(getInstructionCountTo(targets[i])))));
      }
       
      ATerm low  = aint(matches[0]);
      ATerm high = aint(matches[matches.length - 1]);
       
      _result = appl("TABLESWITCH", 
	appl("Count", aint(getInstructionCountTo(instr.getTarget()))),
	appl("Low", low),
	appl("High", high),
	jumps
      );
    }

    /**
     * Misc.
     */

    public void visitBREAKPOINT(BREAKPOINT obj)  
    {       
      _result = appl("BREAKPOINT");
    }

    public void visitIMPDEP1(IMPDEP1 obj)
    {       
      _result = appl("IMPDEP1");
    }
  
    public void visitIMPDEP2(IMPDEP2 obj)
    {       
      _result = appl("IMPDEP2");
    }

    public void visitNOP(NOP obj)  
    {       
      _result = appl("NOP");
    }
  }
}
